import unittest
import pytest

import numpy

import cupy
from cupy import random
from cupy import testing


class BitGeneratorTestCase:

    def setUp(self):
        self.seed = testing.generate_seed()

    def check_seed(self, seed):
        bg1 = self.bg(seed)
        bg2 = self.bg(seed)
        bg3 = self.bg(None)

        xs1 = bg1.random_raw(10)
        xs2 = bg2.random_raw(10)
        xs3 = bg3.random_raw(10)

        # Random state must be reproducible
        assert cupy.array_equal(xs1, xs2)
        # Random state must be initialized randomly with seed=None
        assert not cupy.array_equal(xs1, xs3)

    @testing.for_int_dtypes(no_bool=True)
    def test_seed_not_none(self, dtype):
        self.check_seed(dtype(0))

    @testing.for_dtypes([numpy.complex_])
    def test_seed_invalid_type_complex(self, dtype):
        with self.assertRaises(TypeError):
            self.bg(dtype(0))

    @testing.for_float_dtypes()
    def test_seed_invalid_type_float(self, dtype):
        with self.assertRaises(TypeError):
            self.bg(dtype(0))

    def test_array_seed(self):
        self.check_seed(numpy.random.randint(0, 2**31, size=10))


@testing.with_requires('numpy>=1.17.0')
@testing.fix_random()
@testing.gpu
@pytest.mark.skipif(cupy.cuda.runtime.is_hip,
                    reason='HIP does not support this')
class TestBitGeneratorXORWOW(BitGeneratorTestCase, unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.bg = random._bit_generator.XORWOW


@testing.with_requires('numpy>=1.17.0')
@testing.fix_random()
@testing.gpu
@pytest.mark.skipif(cupy.cuda.runtime.is_hip,
                    reason='HIP does not support this')
class TestBitGeneratorMRG32k3a(BitGeneratorTestCase, unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.bg = random._bit_generator.MRG32k3a


@testing.with_requires('numpy>=1.17.0')
@testing.fix_random()
@testing.gpu
@pytest.mark.skipif(cupy.cuda.runtime.is_hip,
                    reason='HIP does not support this')
class TestBitGeneratorPhilox4x3210(BitGeneratorTestCase, unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.bg = random._bit_generator.Philox4x3210



import unittest

import numpy
import pytest

import cupy
from cupy import testing


@testing.gpu
class TestPolynomial(unittest.TestCase):

    @testing.for_all_dtypes(no_float16=True)
    @testing.numpy_cupy_allclose(rtol=1e-5)
    def test_polyvander1(self, xp, dtype):
        a = testing.shaped_random((10,), xp, dtype)
        return xp.polynomial.polynomial.polyvander(a, 20)

    @testing.for_all_dtypes(no_float16=True)
    @testing.numpy_cupy_allclose(rtol=1e-5)
    def test_polyvander2(self, xp, dtype):
        a = testing.shaped_random((10,), xp, dtype)
        return xp.polynomial.polynomial.polyvander(a, 10)

    @testing.for_all_dtypes(no_float16=True)
    @testing.numpy_cupy_allclose(rtol=1e-5)
    def test_polyvander3(self, xp, dtype):
        a = testing.shaped_random((100,), xp, dtype)
        return xp.polynomial.polynomial.polyvander(a, 10)

    @testing.for_all_dtypes()
    def test_polyvander_negative_degree(self, dtype):
        for xp in (numpy, cupy):
            a = testing.shaped_random((10,), xp, dtype)
            with pytest.raises(ValueError):
                xp.polynomial.polynomial.polyvander(a, -3)

    @testing.with_requires('numpy>=1.17')
    @testing.for_all_dtypes()
    def test_polyvander_non_integral_float_degree(self, dtype):
        for xp in (numpy, cupy):
            a = testing.shaped_random((10,), xp, dtype)
            with pytest.raises(TypeError):
                xp.polynomial.polynomial.polyvander(a, 2.6)

    @testing.with_requires('numpy>=1.17')
    @testing.for_all_dtypes(no_float16=True)
    @testing.numpy_cupy_allclose(rtol=1e-5)
    def test_polyvander_integral_float_degree(self, xp, dtype):
        a = testing.shaped_random((10,), xp, dtype)
        with testing.assert_warns(DeprecationWarning):
            return xp.polynomial.polynomial.polyvander(a, 5.0)

    @testing.for_all_dtypes()
    @testing.numpy_cupy_array_equal()
    def test_polyvander_zeros(self, xp, dtype):
        a = xp.zeros(10, dtype)
        return xp.polynomial.polynomial.polyvander(a, 10)

    @testing.for_all_dtypes()
    @testing.numpy_cupy_allclose(rtol=1e-5)
    def test_polyvander_ndim(self, xp, dtype):
        a = testing.shaped_random((3, 2, 1), xp, dtype)
        return xp.polynomial.polynomial.polyvander(a, 2)

    @testing.for_all_dtypes()
    @testing.numpy_cupy_allclose(rtol=1e-5)
    def test_polyvander_zero_dim(self, xp, dtype):
        a = testing.shaped_random((), xp, dtype)
        return xp.polynomial.polynomial.polyvander(a, 5)

    @testing.for_all_dtypes(no_bool=True)
    @testing.numpy_cupy_allclose(rtol=1e-6)
    def test_polycompanion(self, xp, dtype):
        a = testing.shaped_random((1000,), xp, dtype)
        return xp.polynomial.polynomial.polycompanion(a)

    @testing.for_all_dtypes()
    def test_polycompanion_zeros(self, dtype):
        for xp in (numpy, cupy):
            a = xp.zeros(10, dtype)
            with pytest.raises(ValueError):
                xp.polynomial.polynomial.polycompanion(a)

    @testing.for_all_dtypes()
    def test_polycompanion_zerosize(self, dtype):
        for xp in (numpy, cupy):
            a = xp.zeros((0,), dtype)
            with pytest.raises(ValueError):
                xp.polynomial.polynomial.polycompanion(a)

    @testing.for_all_dtypes(no_bool=True)
    @testing.numpy_cupy_array_equal()
    def test_polycompanion_trailing_zeros(self, xp, dtype):
        a = xp.array([3, 5, 7, 0, 0, 0], dtype)
        return xp.polynomial.polynomial.polycompanion(a)

    @testing.for_all_dtypes()
    def test_polycompanion_single_value1(self, dtype):
        for xp in (numpy, cupy):
            a = xp.array([3, 0, 0, 0], dtype)
            with pytest.raises(ValueError):
                xp.polynomial.polynomial.polycompanion(a)

    @testing.for_all_dtypes(no_bool=True)
    @testing.numpy_cupy_array_equal()
    def test_polycompanion_single_value2(self, xp, dtype):
        a = xp.array([0, 0, 0, 3], dtype)
        return xp.polynomial.polynomial.polycompanion(a)

    @testing.for_all_dtypes(no_bool=True)
    @testing.numpy_cupy_array_equal()
    def test_polycompanion_single_value3(self, xp, dtype):
        a = xp.array([0, 3, 0, 0], dtype)
        return xp.polynomial.polynomial.polycompanion(a)

    @testing.for_all_dtypes()
    def test_polycompanion_ndim(self, dtype):
        for xp in (numpy, cupy):
            a = testing.shaped_random((3, 4, 5), xp, dtype)
            with pytest.raises(ValueError):
                xp.polynomial.polynomial.polycompanion(a)

    @testing.for_all_dtypes()
    def test_polycompanion_zero_dim(self, dtype):
        for xp in (numpy, cupy):
            a = testing.shaped_random((), xp, dtype)
            with pytest.raises(ValueError):
                xp.polynomial.polynomial.polycompanion(a)

    def test_polycompanion_nocommon_types(self):
        for xp in (numpy, cupy):
            a = testing.shaped_random((5,), xp, dtype=bool)
            with pytest.raises(Exception):
                xp.polynomial.polynomial.polycompanion(a)



import re
import unittest

from example_tests import example_test


class TestCustomStruct(unittest.TestCase):

    def test_builtin_vectors(self):
        output = example_test.run_example('custom_struct/builtin_vectors.py')
        assert re.match(
            r"Kernel output matches expected value.",
            output.decode('utf-8'),
        )

    def test_packed_matrix(self):
        output = example_test.run_example('custom_struct/packed_matrix.py')
        assert re.match(
            r"Kernel output matches expected value for type 'float'.\r?\n"
            r"Kernel output matches expected value for type 'double'.",
            output.decode('utf-8'),
        )

    def test_complex_struct(self):
        output = example_test.run_example('custom_struct/complex_struct.py')
        assert re.match(
            r"Overall structure itemsize: \d+ bytes\r?\n"
            r"Structure members itemsize: \[(\s*\d+){5}]\r?\n"
            r"Structure members offsets: \[(\s*\d+){5}]\r?\n"
            r"Complex structure value:\r?\n"
            r"\s+\[.*\]\r?\n"
            r"Kernel output matches expected value.",
            output.decode('utf-8'),
        )
